import java.io.*;
import java.util.*;

import sv.kernel.*;
import sv.script.*;

public class ScriptHandler extends Thread {
  svserver svServer;
  ScriptFrame parent;

  int blockSize, stmtSize;
  int currBlockPos = 0;
  int currStmtsPos = 0;

  StatementBlock[] blocks;
  Statement[] stmts;
  Vector varpool;

  private boolean verbose = true;
      
  public ScriptHandler(svserver server, ScriptFrame frame, String file) {
     this.svServer = server;
     this.parent = frame;

     varpool = new Vector();
     Argument argmt = new Argument(Argument.VARIABLE, "errlvl");
     argmt.setIntVal(-1);
     varpool.addElement(argmt);

     try {
        BufferedReader in
             = new BufferedReader(new FileReader(file));
	parent.append("\n\nParsing: " + file);

        SVParser parser = new SVParser(in);

        Vector vec = parser.getBlocks();
        blocks = new StatementBlock[vec.size()];
        blockSize = vec.size();    
        vec.copyInto(blocks);

        parent.append("  ... done");

     } catch (Exception e) {
        e.printStackTrace();
     }
  }

  /**
   * set verbose mode
   */
  public void setVerboseOn(boolean b) {
     verbose = b;
  }

  /**
   * get verbose mode
   */
  public boolean getVerboseOn() {
     return verbose;
  }

  public void run() {
    suspend();
    parent.append("\nScript starts:");

    do {
        if (verbose)
           parent.append("\n> " + blocks[currBlockPos].getBlockName() +":");

        Vector vec = blocks[currBlockPos].getStatementVector();
        stmtSize = vec.size();
        stmts = new Statement[vec.size()]; 
        vec.copyInto(stmts);

        currBlockPos++;
        currStmtsPos = 0;

        do {              

           Statement stmt = stmts[currStmtsPos];
           int stmttype = stmt.getKeyword();
           Argument[] args = stmt.getArgs();

           if (verbose)
              parent.append("\n> " + stmts[currStmtsPos].getInputString());

           ///////////////////////////////////////////////////////
           if (stmttype == Statement.EXIT) {
              ending();
              return;
           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.WAIT) {
              if (args[0].getType() == Argument.CONSTANT_INT) {
                 try {
                    sleep(args[0].getIntVal());
                 } catch (InterruptedException e) {
                    parent.append("InterruptedException occured.");
                    ending();
                    return;
                 }
              } else {
                 parent.append("\nInterruption. Click \"Continue\" button.");
                 parent.contBtn.setEnabled(true);
                 suspend();
	      }

           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.GOTO) {
              boolean match = false;
              int matched = -1;
              for (int i=0; i<blockSize; i++) {
		      if (blocks[i].getBlockName().equals(args[0].getStringVal())) {
                   match = true;
                   matched = i;
                   break;
                }
              }

              if (match) {
                 currBlockPos = matched;
                 break;   
              } else {
                 parent.append("\nRuntime error: Block mismatch.");
                 ending();
                 return; 
              }

           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.PRINT) {
              if (args[0].getStringVal().equals("console")) {
                 parent.append("\n" + args[1].getStringVal());
              } else {
System.out.println("Not support yet ...");
              }

           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.READ) {
              Argument argmt = 
                   new Argument(Argument.VARIABLE, args[0].getVarName());
              argmt.setStringVal(args[1].getStringVal());

              for (int i=0; i<varpool.size(); i++) {
                 Argument art = (Argument)varpool.elementAt(i); 
                 if (art.getVarName().equals(args[0].getVarName())) 
                    varpool.removeElementAt(i);
                 
              }

              DataInputStream in = null;
              String fname = args[1].getStringVal();
              try {
                 in = new DataInputStream(
       new BufferedInputStream(new FileInputStream(fname)));
              } catch (Exception E) {
                  parent.append("\nCan't find file " + fname + ".");
                  ending();
                  return;
              }
              long fid = SVFileReader.readFromInput(svServer, in, 1, fname);
              if (fid < 0) {
                  parent.append("\nReading error: " + fname);
                  ending();
                  return;
              }

              argmt.setLongVal(fid);
              varpool.addElement(argmt);
              setErrorLvl(0);

	      //System.out.println(args[0].getVarName());
	      //System.out.println(args[1].getStringVal());

           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.IF) {
              int intVal = -1;
              boolean match = false;

              if (args[2].getType() != Argument.CONSTANT_INT) {
                 parent.append("\nVariable comparison is NOT supported yet.");
                 ending();
                 return;
              }
         
              for (int i=0; i<varpool.size(); i++) {
                 Argument art = (Argument)varpool.elementAt(i); 
                 if (art.getVarName().equals(args[0].getVarName())) {
                    match = true;
                    intVal = art.getIntVal();
                    break;
		 }
              }

              if (match) {
                 match = false;

  		 if (args[1].getType() == Argument.EQUAL_OPER) {
                    if (intVal == args[2].getIntVal())
                       match = true;
                 } else if (args[1].getType() == Argument.NOT_EQUAL_OPER) {
                    if (intVal != args[2].getIntVal())
                       match = true;
                 }
              } else {
                 parent.append("\nRuntime error: Variable mismatch.");
                 ending();
                 return; 
              }

              if (match) {
                 match = false;
                 int matched = -1;
                 for (int i=0; i<blockSize; i++) {
		   if (blocks[i].getBlockName().equals(args[3].getStringVal())) {
                      match = true;
                      matched = i;
                      break;
                   }
                 }

                 if (match) {
                    currBlockPos = matched;
                    break;   
                 } else {
                    parent.append("\nRuntime error: Block mismatch.");
                    ending();
                    return; 
                 }
              }

           ///////////////////////////////////////////////////////
           } else if (stmttype == Statement.COMMAND) {
              long dataid = -1;
              boolean match = false;

	      //              System.out.println(args.length);
	      //              System.out.println(args[0].getVarName());

              if (args[0].getType() != Argument.VARIABLE) {
                 parent.append("\nExpected data set variable.");
                 ending();
                 return;
              }

              for (int i=0; i<varpool.size(); i++) {
                 Argument art = (Argument)varpool.elementAt(i); 
                 if (art.getVarName().equals(args[0].getVarName())) {
                    match = true;
                    dataid = art.getLongVal();
                    break;
		 }
              }

              if (match) {
                 match = false;

                 String svcommand = args[1].getStringVal();
                 if (svcommand.equalsIgnoreCase("animate")) {
                    doit(dataid, CommandSet.ANIMATE);
                 } else if (svcommand.equalsIgnoreCase("stop")) {
		    stopCommand(dataid, args);
                 } else if (svcommand.equalsIgnoreCase("next")) {
                    doit(dataid, CommandSet.NEXT);
                 } else if (svcommand.equalsIgnoreCase("previous")) {
                    doit(dataid, CommandSet.PREVIOUS);
                 } else if (svcommand.equalsIgnoreCase("set")) {
                    setCommand(dataid, args);
                 } else {
                    parent.append("\nUnknown Scivis command:" + svcommand);
                    ending();
                    return;
                 }

              } else {
                 parent.append("\nRuntime error: Variable mismatch.");
                 ending();
                 return; 
              }


           ///////////////////////////////////////////////////////
           } else {
              parent.append("\nRuntime error: " +
                               stmts[currStmtsPos].getInputString());
              ending();
              return;
           }

           currStmtsPos++;

        } while (currStmtsPos < stmtSize);

    } while (currBlockPos < blockSize);
    ending();
  }

  private void setCommand(long id, Argument[] args) {
     GFrame gframe = (GFrame)svServer.fmanager.getGFrame(id);
     int rank = gframe.getRank();
     String cName = args[2].getStringVal();

     String[] newargs;
     boolean yesno = false;

     boolean okCommand = false;

     if (rank == DataSet.ONED) {
        /////////////////////////////////////////////////////
        if (cName.equalsIgnoreCase("background_color")) {
           if (args.length < 6) {
              parent.append("Missing RGB values.");
              ending();
              return;
           } else {
              newargs = new String[3];
              try {
                 newargs[0] = String.valueOf(args[3].getIntVal());
                 newargs[1] = String.valueOf(args[4].getIntVal());
                 newargs[2] = String.valueOf(args[5].getIntVal());
                 svServer.fmanager.action1d((GFPlane)gframe, CommandSet.BACKCOLOR, newargs, true);
              } catch (Exception e) {
                 parent.append("Missing RGB values.");
                 ending();
                 return;
	      }
              okCommand = true;
           }

        } else if (cName.equalsIgnoreCase("connect_point")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action1d((GFPlane)gframe, CSPlane.CONNECTPOINT, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("display_label")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action1d((GFPlane)gframe, CSPlane.DISPLAYLBL, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    

        } else if (cName.equalsIgnoreCase("display_time")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action1d((GFPlane)gframe, CSPlane.DISPLAYTIME, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    

        } else if (cName.equalsIgnoreCase("display_legend")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action1d((GFPlane)gframe, CSPlane.DISPLAYLGD, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    

        } else if (cName.equalsIgnoreCase("display_tickmark")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action1d((GFPlane)gframe, CSPlane.TICKMARK, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
        } 
 
     } else if (rank == DataSet.TWOD) {

        /////////////////////////////////////////////////////
        if (cName.equalsIgnoreCase("background_color")) {
           if (args.length < 6) {
              parent.append("Missing RGB values.");
              ending();
              return;
           } else {
              newargs = new String[3];
              try {
                 newargs[0] = String.valueOf(args[3].getIntVal());
                 newargs[1] = String.valueOf(args[4].getIntVal());
                 newargs[2] = String.valueOf(args[5].getIntVal());
                 svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.BACKCOLOR, newargs, true);
              } catch (Exception e) {
                 parent.append("Missing RGB values.");
                 ending();
                 return;
	      }
              okCommand = true;
           }

        } else if (cName.equalsIgnoreCase("bounding_box")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.BOUNDBOX, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("xy_mesh")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.XYMESH, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("face_grid")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.FACEGRID, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("xy_tick")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.XYTICK, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("z_tick")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.ZTICK, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("display_time")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.DISPLAYTIME, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("wireframe_plot")) {

           try {
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.WIRE, null, true);
           } catch (Exception e) {
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("color_wireframe_plot")) {

           try {
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.WIRECOLOR, null, true);
           } catch (Exception e) {
              ending();
              return;
	   }
           okCommand = true;    
           
        }  else if (cName.equalsIgnoreCase("hidden_surface_removal_plot")) {

           try {
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.HIDDEN, null, true);
           } catch (Exception e) {
              ending();
              return;
	   }
           okCommand = true;    
           
        }  else if (cName.equalsIgnoreCase("color_mapped_plot")) {

           try {
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.COLOR, null, true);
           } catch (Exception e) {
              ending();
              return;
	   }
           okCommand = true;    
           
        }  else if (cName.equalsIgnoreCase("point_plot")) {

           try {
              svServer.fmanager.action2d((GFrame2D)gframe, CommandSet2D.POINTPLOT, null, true);
           } catch (Exception e) {
              ending();
              return;
	   }
           okCommand = true;    
           
        }  

     } else if (rank == DataSet.THREED) {

        if (cName.equalsIgnoreCase("bounding_box")) {
           newargs = new String[1];

           try {
              newargs[0] = args[3].getStringVal();
              svServer.fmanager.action3d((GFrame3D)gframe, CommandSet3D.BOUNDBOX, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("wireframe_plot")) {
           newargs = new String[1];

           try {
              newargs[0] = "true";
              svServer.fmanager.action3d((GFrame3D)gframe, CommandSet3D.BOUNDBOX, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("fill_plot")) {
           newargs = new String[1];

           try {
              newargs[0] = "true";
              svServer.fmanager.action3d((GFrame3D)gframe, CommandSet3D.FILL, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } else if (cName.equalsIgnoreCase("shading_plot")) {
           newargs = new String[1];

           try {
              newargs[0] = "true";
              svServer.fmanager.action3d((GFrame3D)gframe, CommandSet3D.SHADING, newargs, true);
           } catch (Exception e) {
              parent.append("Missing flag.");
              ending();
              return;
	   }
           okCommand = true;    
           
        } 

     }

     if (!okCommand) {
        parent.append("Unknown command: " + cName);
        ending();
        return;
     }
  }

  // stop command
  private void stopCommand(long id, Argument[] args) {
     GFrame gframe = (GFrame)svServer.fmanager.getGFrame(id);
     int rank = gframe.getRank();

     String time[] = new String[1];
     time[0] = String.valueOf(args[2].getIntVal());

     if (rank == DataSet.ONED) {
         svServer.fmanager.action1d((GFPlane)gframe, CommandSet.STOP, time, true);
     } else if (rank == DataSet.TWOD) {
         svServer.fmanager.action2d((GFrame2D)gframe, CommandSet.STOP, time, true);
     } else if (rank == DataSet.THREED) {
         svServer.fmanager.action3d((GFrame3D)gframe, CommandSet.STOP, time, true);
     }
  }

  private void doit(long id, int command) {
     GFrame gframe = (GFrame)svServer.fmanager.getGFrame(id);
     int rank = gframe.getRank();

     if (rank == DataSet.ONED) {
         svServer.fmanager.action1d((GFPlane)gframe, command, null, true);
     } else if (rank == DataSet.TWOD) {
         svServer.fmanager.action2d((GFrame2D)gframe, command, null, true);
     } else if (rank == DataSet.THREED) {
         svServer.fmanager.action3d((GFrame3D)gframe, command, null, true);
     }
  }

  private void ending() {
     parent.runBtn.setEnabled(false);
     parent.append("\nScript exits.");
  }


  private void setErrorLvl(int i) {
     Argument argmt = (Argument)varpool.firstElement();     
     argmt.setIntVal(i);
  }

} // end of class

